import csv
from builtins import super

from phigaro.data import (
    read_prodigal_output,
    read_hmmer_output,
    hmmer_res_to_npn,
    hmmer_res_to_gc,
    Gene,
    define_taxonomy,
)
from phigaro.finder.v2 import V2Finder
from phigaro import const
from .base import AbstractTask
from .prodigal import ProdigalTask
from .hmmer import HmmerTask
import os
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from phigaro.to_html.preprocess import plot_html, form_sequence, if_transposable
from phigaro.to_html.html_formation import form_html_document
from phigaro.context import Context
import numpy as np


class RunPhigaroTask(AbstractTask):
    task_name = "run_phigaro"

    def __init__(self, hmmer_task, prodigal_task):
        """
        :type hmmer_task: HmmerTask
        :type prodigal_task: ProdigalTask
        """
        super().__init__()
        self.hmmer_task = hmmer_task
        self.prodigal_task = prodigal_task

    def _prepare(self):
        self.mode = self.config["phigaro"]["mode"]
        self.finder = V2Finder(
            window_len=self.config["phigaro"]["window_len"],
            threshold_min=self.config["phigaro"]["threshold_min_%s" % self.mode],
            threshold_max=self.config["phigaro"]["threshold_max_%s" % self.mode],
            mode=self.mode,
        )
        self._wtp = self.config["phigaro"].get("wtp", False)
        self._print_vogs = self.config["phigaro"].get("print_vogs", False)
        self._filename = self.config["phigaro"].get("filename", False)
        self._no_html = self.config["phigaro"].get("no_html", False)
        self._not_open = self.config["phigaro"].get("not_open", False)
        self._save_fasta = self.config["phigaro"].get("save_fasta", False)
        self._output = self.config["phigaro"].get("output", False)
        if self._wtp:
            self._output_wtp = self.config["phigaro"].get("output_wtp", False)
        self._uuid = self.config["phigaro"].get("uuid", False)
        self.context = Context.instance()

    def output(self):
        return self.file("{}.tsv".format(self.sample))

    def run(self):
        max_evalue = self.config["hmmer"]["e_value_threshold"]
        penalty_black = self.config["phigaro"]["penalty_black"]
        penalty_white = self.config["phigaro"]["penalty_white"]
        gff = self.config["phigaro"]["gff"]
        bed = self.config["phigaro"]["bed"]
        mean_gc = self.config["phigaro"]["mean_gc"]

        pvogs_black_list = const.DEFAULT_BLACK_LIST
        pvogs_white_list = const.DEFAULT_WHITE_LIST

        scaffold_set = read_prodigal_output(self.prodigal_task.output())
        hmmer_result = read_hmmer_output(self.hmmer_task.output())

        gff_base = ["##gff-version 3.2.1"]
        gff_scaffold = []
        gff_prophage = []
        gff_gene = []
        bed_prophage = []
        bed_gene = []
        wtp_output = []

        with open(self.output(), "w") as of:
            writer = csv.writer(of, delimiter="\t")

            if self._print_vogs:
                writer.writerow(
                    (
                        "scaffold",
                        "id",
                        "begin",
                        "end",
                        "transposable",
                        "taxonomy",
                        "vog",
                    )
                )
            else:
                writer.writerow(
                    ("scaffold", "id", "begin", "end", "transposable", "taxonomy")
                )

            plotly_plots = []
            phage_info = []
            transposables_status = []
            for scaffold in scaffold_set:
                phage_num = 0
                if gff:
                    gff_scaffold.append(
                        "##sequence-region %s 1 %d"
                        % (
                            scaffold.name,
                            self.context.scaffolds_info[scaffold.name],
                        )
                    )
                phage_info.append([scaffold.name, []])
                genes = list(scaffold)  # type: list[Gene]
                npn = hmmer_res_to_npn(
                    scaffold,
                    hmmer_result,
                    max_evalue=max_evalue,
                    penalty_black=penalty_black,
                    penalty_white=penalty_white,
                    pvogs_black_list=pvogs_black_list,
                    pvogs_white_list=pvogs_white_list,
                )
                gc = hmmer_res_to_gc(scaffold, hmmer_result, max_evalue=max_evalue)
                if self.mode == "abs":
                    gc = np.array(gc)
                    gc = np.absolute(gc - mean_gc) + mean_gc
                    gc[gc == np.absolute(0 - mean_gc) + mean_gc] = 0
                phages = self.finder.find_phages(npn, gc)

                for phage in phages:
                    phage_num += 1
                    phage_name = "%s_prophage%d" % (scaffold.name, phage_num)
                    begin = genes[phage.begin].begin - 1
                    end = genes[phage.end].end - 1
                    phage_genes = genes[phage.begin : (phage.end + 1)]

                    hmmer_records = [
                        hmmer_result.min_record(
                            hmmer_result.get_records(scaffold.name, gene.name)
                        )
                        for gene in genes[phage.begin : phage.end]
                    ]
                    hmmer_pvogs_records = (
                        record.vog_name
                        for record in hmmer_records
                        if record and record.evalue <= max_evalue
                    )
                    pvogs_records_str = ", ".join(hmmer_pvogs_records)
                    taxonomy = define_taxonomy(pvogs_records_str)
                    hmmer_records = [
                        record
                        for record in hmmer_records
                        if record and record.evalue <= max_evalue
                    ]
                    transposable = if_transposable(hmmer_records)

                    if gff:
                        gff_prophage.append(
                            "\t".join(
                                [
                                    scaffold.name,
                                    ".",
                                    "prophage",
                                    str(begin + 1),
                                    str(end + 1),
                                    ".",
                                    ".",
                                    ".",
                                    "ID=%s;taxonomy=%s;transposable=%s"
                                    % (phage_name, taxonomy, transposable),
                                ]
                            )
                        )
                    if bed:
                        bed_prophage.append(
                            "\t".join(
                                [
                                    scaffold.name,
                                    str(begin),
                                    str(end + 1),
                                    phage_name,
                                    ".",
                                    ".",
                                ]
                            )
                        )

                    if gff or bed:
                        for record in phage_genes:
                            hmmer_record = hmmer_result.min_record(
                                hmmer_result.get_records(scaffold.name, record.name)
                            )
                            pvog = (
                                "." if hmmer_record is None else hmmer_record.vog_name
                            )
                            evalue = (
                                "." if hmmer_record is None else hmmer_record.evalue
                            )
                            if gff:
                                gff_gene.append(
                                    "\t".join(
                                        [
                                            scaffold.name,
                                            ".",
                                            "gene",
                                            str(record.begin),
                                            str(record.end),
                                            str(evalue),
                                            "+" if record.strand > 0 else "-",
                                            ".",
                                            "ID=%s;Parent=%s;pvog=%s"
                                            % (record.name, phage_name, pvog),
                                        ]
                                    )
                                )
                            if bed:
                                bed_gene.append(
                                    "\t".join(
                                        [
                                            scaffold.name,
                                            str(record.begin - 1),
                                            str(record.end),
                                            record.name,
                                            str(evalue),
                                            "+" if record.strand > 0 else "-",
                                        ]
                                    )
                                )

                    if self._print_vogs:
                        writer.writerow(
                            (
                                scaffold.name,
                                phage_name,
                                begin,
                                end,
                                transposable,
                                taxonomy,
                                pvogs_records_str,
                            )
                        )
                    else:
                        writer.writerow(
                            (
                                scaffold.name,
                                phage_name,
                                begin,
                                end,
                                transposable,
                                taxonomy,
                            )
                        )

                    sequence, record = form_sequence(
                        self._filename,
                        "%s_prophage_%d" % (scaffold.name, phage_num),
                        begin,
                        end,
                        scaffold.name,
                    )
                    transposables_status.append(transposable)
                    the_phage_info = [
                        begin,
                        end,
                        taxonomy,
                        sequence,
                        pvogs_records_str,
                    ]
                    phage_info[-1][1].append(the_phage_info)
                    if self._save_fasta:
                        with open(self._output + ".fasta", "a") as f:
                            SeqIO.write(
                                SeqRecord(
                                    record,
                                    id=phage_name,
                                    description="%s_prophage_%d"
                                    % (scaffold.name, phage_num),
                                ),
                                f,
                                "fasta",
                            )
                    if (not self._no_html) and (self._output):
                        plotly_plots.append(plot_html(hmmer_records, begin, end))

                phage_info = (
                    phage_info if len(phage_info[-1][1]) > 0 else phage_info[:-1]
                )

            wtp_output = [the_phage_info[0] for the_phage_info in phage_info]

            if self._wtp:
                with open(self._output_wtp, "w") as f:
                    f.write("\n".join(wtp_output))
            if gff:
                with open(self._output + ".gff3", "w") as f:
                    f.write(
                        "\n".join(gff_base + gff_scaffold + gff_prophage + gff_gene)
                    )
            if bed:
                with open(self._output + ".bed", "w") as f:
                    f.write("\n".join(gff_base + bed_prophage + bed_gene))

            if (len(phage_info) > 0) and (not self._no_html) and (self._output):
                html = form_html_document(
                    phage_info,
                    transposables_status,
                    plotly_plots,
                    self._filename,
                    self._uuid,
                )
                with open(self._output + ".html", "w") as f:
                    f.write(html)
                if not self._not_open:
                    os.system(
                        "xdg-open "
                        + self._output
                        + ".html"
                        + " > /dev/null 2>/dev/null"
                    )



from builtins import super

from .base import AbstractTask
import sh


class ProdigalTask(AbstractTask):
    task_name = "prodigal"

    def __init__(self, preprocess_task):
        super().__init__()

        self.preprocess_task = preprocess_task
        self.prodigal = sh.Command(self.config["prodigal"]["bin"]).bake(
            i=self.preprocess_task.output(), a=self.output()
        )

    def output(self):
        return self.file("{}.faa".format(self.sample))

    def run(self):
        self.prodigal()



from .html_templates import *
from bs4 import BeautifulSoup


def form_html_document(
    prophage_data, transposables_status, plots_data, filename_input, uuid
):
    def form_tbody(data, transposables_data):
        prophage_index = 1
        tbody = BeautifulSoup(features="lxml").new_tag("tbody")
        for scaffold, scaffold_info in data:
            tr = BeautifulSoup(features="lxml").new_tag(
                "tr", **{"class": "scaffold-name"}
            )
            th = BeautifulSoup(features="lxml").new_tag("th", scope="row", colspan="4")
            th.append(scaffold)
            tr.append(th)
            tbody.append(tr)
            for prophage_info in scaffold_info:
                index_to_paste = (
                    BeautifulSoup(
                        transposable_index.format(prophage_index), "html.parser"
                    )
                    if transposables_data[prophage_index - 1]
                    else prophage_index
                )
                if prophage_index == 1:
                    prophage_info = (
                        [prophage_index] + prophage_info + ["active", index_to_paste]
                    )
                else:
                    prophage_info = (
                        [prophage_index] + prophage_info + ["", index_to_paste]
                    )
                tbody.append(BeautifulSoup(row.format(*prophage_info), "html.parser"))
                prophage_index += 1
        return str(tbody)

    def form_plots_body(data):
        plots_body = BeautifulSoup(features="lxml")
        for prophage_index, plot in enumerate(data):
            if prophage_index == 0:
                div = BeautifulSoup(features="lxml").new_tag(
                    "div",
                    **{
                        "class": "tab-pane fade show active",
                        "id": "phage-%d" % (prophage_index + 1),
                    }
                )
                inner_div = BeautifulSoup(features="lxml").new_tag(
                    "div", **{"class": "prophage-plot"}
                )
            else:
                div = BeautifulSoup(features="lxml").new_tag(
                    "div",
                    **{
                        "class": "tab-pane fade show",
                        "id": "phage-%d" % (prophage_index + 1),
                    }
                )
                inner_div = BeautifulSoup(features="lxml").new_tag(
                    "div", **{"class": "prophage-plot", "style": "display:none"}
                )
            inner_div.append(BeautifulSoup(plot, "html.parser"))
            div.append(inner_div)
            plots_body.append(div)
        return str(plots_body)

    def arrange_html_parts(
        prophage_data, transposables_status, plots_data, filename_input, uuid
    ):
        global header, style_css, body_main, body_table, body_plots, row, footer
        header = header.format(filename_input)
        tbody = form_tbody(prophage_data, transposables_status)
        plots_body = form_plots_body(plots_data)
        body_table = body_table.format(tbody)
        body_plots = body_plots.format(plots_body)
        body_main = body_main.format(body_table, body_plots)
        footer = footer.replace(
            '<script type="text/javascript">',
            '<script type="text/javascript">\n    var uuid="%s";' % uuid,
        )
        html = header + style_css + body_main + footer
        return html

    return arrange_html_parts(
        prophage_data, transposables_status, plots_data, filename_input, uuid
    )



from builtins import super

import os

from .base import AbstractTask
import sh


class GeneMarkTask(AbstractTask):
    task_name = "gene_mark"

    def __init__(self, input):
        super().__init__()

        self.input = input
        self._lst_file = input + ".lst"
        self.gene_mark = sh.Command(self.config["genemark"]["bin"]).bake(
            m=self.config["genemark"]["mod_path"],
            A=self.output(),
        )

    def output(self):
        return self.file("{}.faa".format(self.sample))

    def run(self):
        self.gene_mark(self.input)

    def clean(self):
        os.unlink(self._lst_file)



from builtins import super

import sh

from .base import AbstractTask
from .prodigal import ProdigalTask


class HmmerTask(AbstractTask):
    task_name = "hmmer"

    def __init__(self, prodigal_task):
        """

        :type prodigal_task: ProdigalTask
        """
        super().__init__()
        self.prodigal = prodigal_task

    def _prepare(self):
        self.hmmer = sh.Command(self.config["hmmer"]["bin"])

    def output(self):
        return self.file("{}.hmmer_out".format(self.sample))

    def run(self):
        self.hmmer(
            "--cpu",
            self.context.threads,
            "--notextw",
            "--tblout",
            self.output(),
            self.config["hmmer"]["pvog_path"],
            self.prodigal.output(),
        )



from __future__ import absolute_import

import argparse
import logging
import multiprocessing
import os
import sys
import uuid
from os.path import join, exists

import yaml

from phigaro.context import Context
from phigaro.batch.runner import run_tasks_chain
from phigaro.batch.task.path import sample_name
from phigaro.batch.task.prodigal import ProdigalTask
from phigaro.batch.task.hmmer import HmmerTask
from phigaro.batch.task.dummy import DummyTask
from phigaro.batch.task.preprocess import PreprocessTask
from phigaro.batch.task.run_phigaro import RunPhigaroTask
from phigaro._version import __version__


def parse_substitute_output(subs):
    subs = subs or []
    res = {}
    for sub in subs:
        task_name, output = sub.split(":")
        res[task_name] = DummyTask(output, task_name)
    return res


def create_task(substitutions, task_class, *args, **kwargs):
    # TODO: refactor to class Application
    task = task_class(*args, **kwargs)
    if task.task_name in substitutions:
        print(
            "Substituting output for {}: {}".format(
                task.task_name, substitutions[task.task_name].output()
            )
        )

        return substitutions[task.task_name]
    return task


def clean_fold():
    is_empty = True
    for root, dirs, files in os.walk("proc", topdown=False):
        for name in files:
            is_empty = False
            break
        if is_empty:
            for name in dirs:
                os.rmdir(os.path.join(root, name))
    if is_empty:
        os.rmdir("proc")


def main():
    default_config_path = join(os.getenv("HOME"), ".phigaro", "config.yml")
    parser = argparse.ArgumentParser(
        prog="phigaro",
        description="Phigaro is a scalable command-line tool for predictions phages and prophages "
        "from nucleid acid sequences",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    parser.add_argument(
        "-f",
        "--fasta-file",
        help="Assembly scaffolds/contigs or full genomes, required",
        required=True,
    )
    parser.add_argument(
        "-c",
        "--config",
        default=default_config_path,
        help="Path to the config file, not required. The deafult is %s"
        % default_config_path,
    )
    parser.add_argument("-v", "--verbose", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "-p",
        "--print-vogs",
        help="Print phage vogs for each region",
        action="store_true",
    )
    parser.add_argument(
        "-e",
        "--extension",
        default=["html"],
        nargs="+",
        help="Type of the output: html, tsv, gff, bed or stdout. Default is html. You can specify several file formats with a space as a separator. Example: -e tsv html stdout.",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="",
        help="Output filename for html and txt outputs. Required by default, but not required for stdout only output.",
    )
    parser.add_argument(
        "--not-open",
        help="Do not open html file automatically, if html output type is specified.",
        action="store_true",
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=multiprocessing.cpu_count(),
        help="Num of threads ("
        "default is num of CPUs={})".format(multiprocessing.cpu_count()),
    )
    parser.add_argument(
        "--no-cleanup",
        action="store_true",
        help="Do not delete any temporary files that was generated by Phigaro (HMMER & Prodigal outputs and some others).",
    )
    parser.add_argument(
        "-S",
        "--substitute-output",
        action="append",
        help="If you have precomputed prodigal and/or hmmer data you can provide paths to the files in the following format: program:address/to/the/file. In place of program you should write hmmer or prodigal. If you need to provide both files you should pass them separetely as two parametres.",
    )
    parser.add_argument(
        "--save-fasta",
        action="store_true",
        help="Save all phage fasta sequences in a fasta file.",
    )
    parser.add_argument(
        "-d",
        "--delete-shorts",
        action="store_true",
        help="Exclude sequences with length < 20000 automatically.",
    )
    parser.add_argument(
        "-m",
        "--mode",
        default="basic",
        help="You can launch Phigaro at one of 3 modes: basic, abs, without_gc. Default is basic. Read more about modes at https://github.com/bobeobibo/phigaro/",
    )
    parser.add_argument("--wtp", action="store_true", help=argparse.SUPPRESS)

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if args.verbose else logging.WARN)
    logging.getLogger("sh.command").setLevel(logging.WARN)

    logger = logging.getLogger(__name__)

    if not exists(args.config):
        # TODO: pretty message
        print("Please, create config file using phigaro-setup script")
        exit(1)

    args.extension = [atype.lower() for atype in args.extension]
    for ext in args.extension:
        if ext not in ["html", "gff", "bed", "tsv", "stdout"]:
            print(
                "Error! The unknown output format in -e/--extensionn parameter: %s. Please, choose one or several from the list: html, gff, bed, tsv, stdout"
                % ext
            )
            exit(1)
    if (args.output == "") and (args.extension != ["stdout"]):
        print(
            "Error! Argument -o/--output is required or change the type of the output to stdout."
        )
        exit(1)

    with open(args.config) as f:
        logger.info("Using config file: {}".format(args.config))
        config = yaml.load(f, Loader=yaml.FullLoader)

    config["phigaro"]["wtp"] = args.wtp
    config["phigaro"]["print_vogs"] = args.print_vogs
    config["phigaro"]["filename"] = args.fasta_file
    config["phigaro"]["no_html"] = True if "html" not in args.extension else False
    config["phigaro"]["not_open"] = args.not_open
    config["phigaro"]["output"] = (
        args.output
        + "/"
        + os.path.splitext(os.path.basename(args.fasta_file))[0]
        + ".phigaro"
    ).replace("//", "/")
    config["phigaro"]["uuid"] = uuid.uuid4().hex
    config["phigaro"]["delete_shorts"] = args.delete_shorts
    config["phigaro"]["gff"] = True if ("gff" in args.extension) else False
    config["phigaro"]["bed"] = True if ("bed" in args.extension) else False
    config["phigaro"]["mode"] = args.mode
    config["phigaro"]["save_fasta"] = args.save_fasta

    filename = args.fasta_file
    sample = "{}-{}".format(sample_name(filename), config["phigaro"]["uuid"])

    if args.wtp:
        config["phigaro"]["not_open"] = True
        config["phigaro"]["gff"] = True
        config["phigaro"]["bed"] = True
        args.extension.append("tsv")
        config["phigaro"]["delete_shorts"] = True
        config["phigaro"]["print_vogs"] = True
        config["phigaro"]["output_wtp"] = args.output + "/phigaro.txt"
        config["phigaro"]["output"] = args.output + "/phigaro/phigaro"
        config["phigaro"]["save_fasta"] = True

    if config["phigaro"]["output"] != "":
        fold = os.path.dirname(config["phigaro"]["output"])
        if fold and not os.path.isdir(fold):
            os.makedirs(fold)
        if args.wtp:
            fold = os.path.dirname(config["phigaro"]["output_wtp"])
            if fold and not os.path.isdir(fold):
                os.makedirs(fold)

    Context.initialize(
        sample=sample,
        config=config,
        threads=args.threads,
    )

    substitutions = parse_substitute_output(args.substitute_output)

    preprocess_task = create_task(substitutions, PreprocessTask, filename)

    prodigal_task = create_task(
        substitutions, ProdigalTask, preprocess_task=preprocess_task
    )
    hmmer_task = create_task(substitutions, HmmerTask, prodigal_task=prodigal_task)

    run_phigaro_task = create_task(
        substitutions,
        RunPhigaroTask,
        prodigal_task=prodigal_task,
        hmmer_task=hmmer_task,
    )

    tasks = [preprocess_task, prodigal_task, hmmer_task, run_phigaro_task]
    task_output_file = run_tasks_chain(tasks)

    if ("tsv" in args.extension) or ("stdout" in args.extension):
        with open(task_output_file) as f:
            f = list(f)
            if "tsv" in args.extension:
                out_f = open(config["phigaro"]["output"] + ".tsv", "w")
                for line in f:
                    out_f.write(line)
            if "stdout" in args.extension:
                out_f = sys.stdout
                for line in f:
                    out_f.write(line)
                out_f.close()

    if not args.no_cleanup:
        for t in tasks:
            t.clean()
        clean_fold()


if __name__ == "__main__":
    main()



from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), "r") as fh:
    long_description = fh.read()

setup(
    name="phigaro",
    description="Phigaro is a scalable command-line tool for predicting phages and prophages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="2.4.0",
    license="MIT",
    author="E.Starikova, N.Pryanichnikov, P.Tikhonova",
    author_email="hed.robin@gmail.com",
    url="https://github.com/bobeobibo/phigaro",
    packages=[
        "phigaro",
        "phigaro.finder",
        "phigaro.cli",
        "phigaro.batch",
        "phigaro.batch.task",
        "phigaro.misc",
        "phigaro.to_html",
    ],
    package_data={"": ["*.json", "*.csv", "README.md"]},
    entry_points={
        "console_scripts": [
            "phigaro = phigaro.cli.batch:main",
            "phigaro-setup = phigaro.cli.helper:main",
        ]
    },
    install_requires=[
        "numpy",
        "six>=1.7.0",
        "pandas>=0.23.4",
        "sh",
        "singleton",
        "PyYAML>=5.1",
        "future",
        "argparse",
        "numpy",
        "plotly",
        "bs4",
        "beautifulsoup4>=4.4.0",
        "lxml",
        "biopython",
    ],
)



from __future__ import absolute_import

import argparse
import csv

import sys
from phigaro import const
from phigaro.finder.base import AbstractFinder, Phage
from phigaro.finder.v2 import V2Finder
from phigaro.data import convert_npn


def find_and_print_phages(npn_filename, finder, sep="\t"):
    # type: (str, AbstractFinder) -> list[(str, str, list[Phage])]

    writer = csv.writer(sys.stdout, delimiter="\t")
    writer.writerow(("scaffold", "begin", "end"))

    with open(npn_filename) as f:
        reader = csv.reader(f, delimiter=sep)
        for scaffold, npn_str in reader:
            npn = convert_npn(npn_str, "P")
            phages = finder.find_phages(npn)
            for phage in phages:
                writer.writerow((scaffold, phage.begin, phage.end))


def main():
    p = argparse.ArgumentParser(
        description="Tool for predicting prophages in metagenomes"
    )
    p.add_argument("npn_file")
    p.add_argument("-w", "--window-len", default=const.DEFAULT_WINDOW_SIZE)
    p.add_argument("-m", "--threshold-min", default=const.DEFAULT_THRESHOLD_MIN)
    p.add_argument("-M", "--threshold-max", default=const.DEFAULT_THRESHOLD_MAX)

    args = p.parse_args()

    finder = V2Finder(
        window_len=args.window_len,
        threshold_min=args.threshold_min,
        threshold_max=args.threshold_max,
    )

    find_and_print_phages(args.npn_file, finder)


if __name__ == "__main__":
    main()



import pandas as pd
import plotly.graph_objs as go
from plotly.offline import plot
import numpy as np
import os
from Bio import SeqIO

pvogs_groups = [
    "Lysis",
    "Integration",
    "Terminase",
    "Replication",
    "Coat",
    "Baseplate",
    "Tail",
    "Assembly",
    "Portal",
    "Other (structural)",
    "Other",
]
cmap = [
    (31, 119, 180),
    (255, 127, 14),
    (44, 160, 44),
    (214, 39, 40),
    (188, 189, 34),
    (23, 190, 207),
    (148, 103, 189),
    (227, 119, 194),
    (0, 0, 128),
    (32, 50, 50),
    (127, 127, 127),
]
colors = {group: cmap[i] for i, group in enumerate(pvogs_groups)}

annotations = pd.read_csv(
    os.path.dirname(os.path.abspath(__file__)) + "/pvogs_annotations.csv", index_col=0
)


def form_arrow(begin, end, nonreverse, width):
    """
    :param begin: int (of gene)
    :param end: int (of gene)
    :param nonreverse: boolean
    :param width:
    :return: tuple of tuples (arrow coords)
    """
    k = 0.05
    height = 0.9 if width < 0.4 else 1
    width = width if width > 0.6 else 0.6
    if nonreverse:
        return (
            (begin, 1 - width * k),
            (begin, 1 + width * k),
            (begin + 0.6 * (end - begin), 1 + width * k),
            (begin + 0.6 * (end - begin), 1 + 1.8 * height * k),
            (end, 1),
            (begin + 0.6 * (end - begin), 1 - 1.8 * height * k),
            (begin + 0.6 * (end - begin), 1 - width * k),
        )
    else:
        return (
            (end, 1 - width * k),
            (end, 1 + width * k),
            (end - 0.4 * (end - begin), 1 + width * k),
            (end - 0.4 * (end - begin), 1 + 1.8 * height * k),
            (begin, 1),
            (end - 0.4 * (end - begin), 1 - 1.8 * height * k),
            (end - 0.4 * (end - begin), 1 - width * k),
        )


def plot_html(records, prophage_begin, prophage_end):
    """
    :param records: list of Gene
    :return: html_plot
    """

    def _plotly_to_html_(
        prophage_begin, prophage_end, func_groups, widths, coords, arrow_coords
    ):
        global colors, pvogs_groups
        traces0 = [
            go.Scatter(
                x=[prophage_begin],
                y=[100],
                mode="markers",
                opacity=1,
                hoverinfo="none",
                marker=dict(
                    size=10,
                    color="rgba(%.5f, %.5f, %.5f, 0.8)" % colors[group],
                ),
                name=group,
            )
            for group in pvogs_groups
            if group in func_groups.keys()
        ] + [
            go.Scatter(
                x=np.array(coords)[elements],
                y=[1] * len(elements),
                text=np.array(records_info)[elements],
                hoverinfo="text",
                mode="markers",
                marker=dict(
                    size=90 * np.array(widths)[elements],
                    color="rgba(%.5f, %.5f, %.5f, 1)" % colors[group],
                    opacity=0,
                ),
                showlegend=False,
            )
            for group, elements in func_groups.items()
        ]
        trace1 = [
            go.Scatter(
                x=[prophage_begin, prophage_end],
                y=[1, 1],
                mode="lines",
                line=dict(color="rgba(116,116,116, 0.5)", width=1),
                hoverinfo="none",
                showlegend=False,
            )
        ]
        layout = {
            "template": "plotly_white",
            "yaxis": dict(
                range=[-0.5, 2.5],
                showgrid=False,
                zeroline=False,
                showline=False,
                ticks="",
                showticklabels=False,
            ),
            "xaxis": dict(title="Nucleotide Number"),
            "shapes": [
                {
                    "type": "path",
                    "path": "M "
                    + " ".join(
                        [
                            "L {:.5f},{:.5f}".format(*arrow_point)
                            for arrow_point in arrow
                        ]
                    )[1:]
                    + " Z",
                    "fillcolor": "rgba(%.5f, %.5f, %.5f, 1)"
                    % colors[group],  # 'rgba(44, 160, 101, 0.5)'
                    "opacity": 0.7,
                    "line": {
                        "color": "rgba(%.5f, %.5f, %.5f, 0.7)" % colors[group],
                    },
                }
                for group, elements in func_groups.items()
                for arrow in np.array(arrow_coords)[elements]
            ],
        }
        data = traces0 + trace1
        fig = {
            "data": data,
            "layout": layout,
        }
        return fig

    pvogs = []
    widths = []
    coords = []
    arrow_coords = []
    records_info = []
    func_groups = {}
    for i, record in enumerate(records):
        pvogs.append(record.vog_name)
        widths.append(record.end - record.begin)
        coords.append(record.begin + (record.end - record.begin) / 2.0)
        arrow_coords.append((record.begin, record.end, (record.strand + 1) / 2))
        records_info.append(
            "%s<br>%d-%d<br>gc_cont = %s"
            % (record.vog_name, record.begin, record.end, record.gc_cont)
        )
        group = (
            annotations.loc[record.vog_name.strip()].group
            if record.vog_name.strip() in annotations.index
            else "Other"
        )
        if group in func_groups.keys():
            func_groups[group].append(i)
        else:
            func_groups[group] = [i]
    if len(widths) > 0:
        widths = 1.0 * (np.array(widths) - min(widths)) / max(widths) + 0.1
        arrow_coords = [
            form_arrow(*info, width=width) for info, width in zip(arrow_coords, widths)
        ]
    plotly_fig = _plotly_to_html_(
        prophage_begin, prophage_end, func_groups, widths, coords, arrow_coords
    )
    return plot(
        plotly_fig,
        include_plotlyjs=False,
        filename="gene_map.html",
        output_type="div",
    )


def form_sequence(fasta_file, fasta_name, prophage_begin, prophage_end, scaffold_name):
    for record in SeqIO.parse(fasta_file, "fasta"):
        if record.id.strip() == scaffold_name.strip():
            sequence = record.seq[prophage_begin : (prophage_end + 1)]
            clean_sequence = sequence
            ins_ind = 77
            while ins_ind < len(sequence):
                sequence = sequence[:ins_ind] + "%0A" + sequence[ins_ind:]
                ins_ind += 80
            sequence = "%" + "3E%s" % fasta_name + "%" + "0A%s" % sequence
            return sequence, clean_sequence


def if_transposable(records):
    status = False
    for record in records:
        group = (
            annotations.loc[record.vog_name.strip()].group
            if record.vog_name.strip() in annotations.index
            else "Other"
        )
        if (group == "Integration") & status:
            return True
        elif group == "Integration":
            status = True
        elif (group != "Integration") & (group != "Other"):
            status = False
    return False



from builtins import input
import os
from os.path import join, exists, isfile
import sh
from future.backports.urllib.parse import urljoin
import re

from phigaro import const


class HelperException(Exception):
    message = None


class ProdigalNotFound(HelperException):
    message = (
        "Didn't find Prodigal anywhere, "
        + "please, download it from https://github.com/hyattpd/Prodigal/wiki/installation if you haven't done this before..."
    )


class HMMERNotFound(HelperException):
    message = "No HMMER found. Please, download HMMER package from here: http://hmmer.org/download.html if you haven't done this before..."


def _choose_option(message, options):
    options = [s.rstrip() for s in options[:-1] if isfile(s.rstrip())] + [options[-1]]

    if len(options) == 1:
        return options[0]

    print(message)
    while True:
        res = None
        for i, option in enumerate(options):
            print("[{}] {}".format(i + 1, option))
        option_num = input("Choose your option (Enter for {}): ".format(options[0]))
        if option_num == "":
            option_num = "1"
        if re.match(r"^\d+$", option_num):
            if int(option_num) == len(options):
                which_program = "prodigal" if "Prodigal" in message else "hmmsearch"
                res = input("Please, write a path to {}: ".format(which_program))
                return res
            option_num = int(option_num) - 1
            if 0 <= option_num < len(options):
                res = options[option_num]
        if res is None:
            print(
                "Please select valid option number ({} to {})".format(1, len(options))
            )
            continue
        return res


def download_file(filename, base_url, out_dir):
    url = urljoin(base_url, filename)
    out = join(out_dir, filename)
    if exists(out):
        while True:
            overwrite = input("File {} already exists, overwrite it (Y/N)?".format(out))
            if overwrite.upper() in {"Y", "YES"}:
                overwrite = "Y"
                break
            elif overwrite.upper() in {"N", "NO"}:
                overwrite = "N"
                break
        if overwrite == "N":
            return

    print(
        "Downloading {url} to {out}".format(
            url=url,
            out=out,
        )
    )
    sh.wget("-O", out, url, _tty_out=True)


def download_pvogs(base_url, out_dir):
    if not exists(out_dir):
        os.makedirs(out_dir)
    # TODO: refactor 'allpvoghmms' const
    download_file("allpvoghmms", base_url, out_dir)
    download_file("allpvoghmms.h3f", base_url, out_dir)
    download_file("allpvoghmms.h3i", base_url, out_dir)
    download_file("allpvoghmms.h3m", base_url, out_dir)
    download_file("allpvoghmms.h3p", base_url, out_dir)


class SetupHelper(object):
    def __init__(self, dont_update_db):
        self.HOME = os.getenv("HOME")
        self._dont_update_db = dont_update_db

    def _locate(self, *args, **kwargs):
        if not self._dont_update_db:
            try:
                print(
                    "If you do not want to run sudo, use the following command to configure Phigaro:"
                )
                print("       phigaro-setup --no-updatedb")
                with sh.contrib.sudo:
                    sh.updatedb()
                    self._dont_update_db = True
            except sh.ErrorReturnCode_1:
                print("Invalid password")
                exit(1)
        try:
            return [s.rstrip() for s in sh.locate(*args, **kwargs)]
        except (sh.ErrorReturnCode_1, sh.CommandNotFound):
            return []

    def _find_binary(self, name, options_message, raise_if_not_found):
        locations = [sh.which(name) or ""] + self._locate(
            "-b", "-r", "^{}$".format(name)
        )
        locations = [l.rstrip() for l in locations if l.rstrip()]

        if not locations:
            # raise raise_if_not_found()
            print(raise_if_not_found)
            which_program = (
                "prodigal" if "Prodigal" in raise_if_not_found else "hmmsearch"
            )
            locations = input("Write a full path to {}:".format(which_program))
            return locations

        locations.append("Add another path manually")
        return _choose_option(
            message=options_message,
            options=locations,
        )

    def _setup_prodigal(
        self,
    ):
        prodigal_location = self._find_binary(
            name="prodigal",
            options_message="Please select appropriate Prodigal location",
            raise_if_not_found=ProdigalNotFound.message,
        )
        return {
            "bin": prodigal_location,
        }

    def _setup_hmmer(self):
        hmmsearch_location = self._find_binary(
            name="hmmsearch",
            options_message="Please select appropriate HMMER location",
            raise_if_not_found=HMMERNotFound.message,
        )
        return {
            "bin": hmmsearch_location,
        }

    def setup(self):
        prodigal_params = self._setup_prodigal()
        hmmer_params = self._setup_hmmer()

        return {
            "prodigal": {
                "bin": prodigal_params["bin"],
            },
            "hmmer": {
                "bin": hmmer_params["bin"],
                "e_value_threshold": const.DEFAULT_MAX_EVALUE,
            },
            "phigaro": {
                "window_len": const.DEFAULT_WINDOW_SIZE,
                "threshold_min_basic": const.DEFAULT_THRESHOLD_MIN_BASIC,
                "threshold_max_basic": const.DEFAULT_THRESHOLD_MAX_BASIC,
                "threshold_min_abs": const.DEFAULT_THRESHOLD_MIN_ABS,
                "threshold_max_abs": const.DEFAULT_THRESHOLD_MAX_ABS,
                "threshold_min_without_gc": const.DEFAULT_THRESHOLD_MIN_WITHOUT_GC,
                "threshold_max_without_gc": const.DEFAULT_THRESHOLD_MAX_WITHOUT_GC,
                "mean_gc": const.DEFAULT_MEAN_GC,
                "penalty_black": const.DEFAULT_PENALTY_BLACK,
                "penalty_white": const.DEFAULT_PENALTY_WHITE,
            },
        }



from __future__ import absolute_import

import argparse
import os
from builtins import input
import sys
import yaml
from os.path import exists, dirname
import sh
from phigaro.helper import (
    SetupHelper,
    HelperException,
    download_pvogs,
    download_file,
)


def create_config(no_update_db, config_path, pvogs_dir, force):
    config_dir = dirname(config_path)
    if not exists(config_dir):
        os.makedirs(config_dir)

    if exists(config_path) and not force:
        print("Phigaro already configured")
        return

    helper = SetupHelper(no_update_db)
    try:
        config = helper.setup()
        # TODO: refactor 'allpvoghmms' const
        config["hmmer"]["pvog_path"] = os.path.join(pvogs_dir, "allpvoghmms")

        print("Found Prodigal in: {}".format(config["prodigal"]["bin"]))
        print("Found HMMER in: {}".format(config["hmmer"]["bin"]))
        print("HMMER model in: {}".format(config["hmmer"]["pvog_path"]))

        with open(config_path, "w") as f:
            yaml.safe_dump(config, f, default_flow_style=False)
    except HelperException as ex:
        sys.stdout.write(ex.message + "\n")
        exit(1)
        return


def read_config_and_download_pvogs(config_path):
    if not exists(config_path):
        print("You must configure phigaro first")
        exit(1)

    with open(config_path) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
        pvogs_dir = dirname(config["hmmer"]["pvog_path"])
        print("Downloading models to {}".format(pvogs_dir))
        download_pvogs("http://download.ripcm.com/phigaro/", pvogs_dir)


def download_test_data():
    test_data_dir = input(
        "Please, write a full path to the directory you want test data saved to"
        + "(by default - working directory):"
    )
    if test_data_dir == "":
        test_data_dir = os.getcwd()
    test_data_dir = test_data_dir.replace("\\", "/")
    test_data_dir = (
        test_data_dir + "/test_data"
        if (test_data_dir[-1] != "/")
        else test_data_dir + "test_data"
    )
    base_url = "https://cdn.jsdelivr.net/gh/bobeobibo/phigaro/test_data/"
    if not exists(test_data_dir):
        os.makedirs(test_data_dir)
    download_file("Bacillus_anthracis_str_ames.fna", base_url, test_data_dir)
    download_file("Bacillus_anthracis_str_ames.phg", base_url, test_data_dir)
    download_file("Bacillus_anthracis_str_ames.phg.html", base_url, test_data_dir)


def main():
    home = os.getenv("HOME")
    phigaro_dir = os.path.join(home, ".phigaro")
    phigaro_config = os.path.join(phigaro_dir, "config.yml")
    pvogs_dir = os.path.join(phigaro_dir, "pvog")

    parser = argparse.ArgumentParser(
        prog="phigaro-setup",
        description="Phigaro setup helper",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-c",
        "--config",
        default=phigaro_config,
        help="Path to a config.yml, default is {}".format(phigaro_config),
    )
    parser.add_argument(
        "-p",
        "--pvog",
        default=pvogs_dir,
        help="pvogs directory, default is {}".format(pvogs_dir),
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force configuration and rewrite config.yml if exists",
    )
    parser.add_argument(
        "--no-updatedb", action="store_true", help="Do not run sudo updatedb"
    )
    # parser.add_argument('--download-models', action='store_true', help='Skip configuration step and download models')
    args = parser.parse_args()
    if args.force:
        sh.rm("-rf", args.config)
    create_config(
        no_update_db=args.no_updatedb,
        config_path=args.config,
        pvogs_dir=args.pvog,
        force=args.force,
    )

    read_config_and_download_pvogs(args.config)
    while True:
        overwrite = input("Do you want to download test data Y/N?")
        if overwrite.upper() in {"Y", "YES"}:
            overwrite = "Y"
            break
        elif overwrite.upper() in {"N", "NO"}:
            overwrite = "N"
            break
    if overwrite == "Y":
        download_test_data()
        print("To run phigaro on test data enter the following command:")
        print("")
        print(
            "       phigaro -f test_data/Bacillus_anthracis_str_ames.fna -o test_data/Bacillus_anthracis_str_ames.phg -p --not-open"
        )
        print("")

    print(
        "The installation process is finished. If you have any questions, you can visit our github: https://github.com/bobeobibo/phigaro."
    )


if __name__ == "__main__":
    main()
